package org.apache.bcel.verifier.statics;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.apache.bcel.Repository;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.LineNumber;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.verifier.PassVerifier;
import org.apache.bcel.verifier.VerificationResult;
import org.apache.bcel.verifier.Verifier;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.ClassConstraintException;
import org.apache.bcel.verifier.exc.InvalidMethodException;
import org.apache.bcel.verifier.exc.StaticCodeConstraintException;
import org.apache.bcel.verifier.exc.StaticCodeInstructionConstraintException;
import org.apache.bcel.verifier.exc.StaticCodeInstructionOperandConstraintException;
import org.apache.bcel.verifier.statics.IntList;

public final class Pass3aVerifier extends PassVerifier {
  private final Verifier myOwner;
  
  private final int methodNo;
  
  private InstructionList instructionList;
  
  private Code code;
  
  private static boolean contains(int[] ints, int i) {
    for (int k : ints) {
      if (k == i)
        return true; 
    } 
    return false;
  }
  
  public Pass3aVerifier(Verifier owner, int methodNo) {
    this.myOwner = owner;
    this.methodNo = methodNo;
  }
  
  private void delayedPass2Checks() {
    int[] instructionPositions = this.instructionList.getInstructionPositions();
    int codeLength = (this.code.getCode()).length;
    LineNumberTable lnt = this.code.getLineNumberTable();
    if (lnt != null) {
      LineNumber[] lineNumbers = lnt.getLineNumberTable();
      IntList offsets = new IntList();
      for (LineNumber lineNumber : lineNumbers) {
        int arrayOfInt[] = instructionPositions, i = arrayOfInt.length;
        byte b = 0;
        while (true) {
          if (b < i) {
            int instructionPosition = arrayOfInt[b];
            int offset = lineNumber.getStartPC();
            if (instructionPosition == offset) {
              if (offsets.contains(offset)) {
                addMessage("LineNumberTable attribute '" + this.code.getLineNumberTable() + "' refers to the same code offset ('" + offset + "') more than once which is violating the semantics [but is sometimes produced by IBM's 'jikes' compiler].");
                break;
              } 
              offsets.add(offset);
              break;
            } 
            b++;
            continue;
          } 
          throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has a LineNumberTable attribute '" + this.code.getLineNumberTable() + "' referring to a code offset ('" + lineNumber
              .getStartPC() + "') that does not exist.");
        } 
      } 
    } 
    Attribute[] atts = this.code.getAttributes();
    for (Attribute att : atts) {
      if (att instanceof LocalVariableTable)
        ((LocalVariableTable)att).forEach(localVariable -> {
              int startpc = localVariable.getStartPC();
              int length = localVariable.getLength();
              if (!contains(instructionPositions, startpc))
                throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has a LocalVariableTable attribute '" + this.code.getLocalVariableTable() + "' referring to a code offset ('" + startpc + "') that does not exist."); 
              if (!contains(instructionPositions, startpc + length) && startpc + length != codeLength)
                throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has a LocalVariableTable attribute '" + this.code.getLocalVariableTable() + "' referring to a code offset start_pc+length ('" + (startpc + length) + "') that does not exist."); 
            }); 
    } 
    CodeException[] exceptionTable = this.code.getExceptionTable();
    for (CodeException element : exceptionTable) {
      int startpc = element.getStartPC();
      int endpc = element.getEndPC();
      int handlerpc = element.getHandlerPC();
      if (startpc >= endpc)
        throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has its start_pc ('" + startpc + "') not smaller than its end_pc ('" + endpc + "')."); 
      if (!contains(instructionPositions, startpc))
        throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its start_pc ('" + startpc + "')."); 
      if (!contains(instructionPositions, endpc) && endpc != codeLength)
        throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its end_pc ('" + startpc + "') [that is also not equal to code_length ('" + codeLength + "')]."); 
      if (!contains(instructionPositions, handlerpc))
        throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its handler_pc ('" + handlerpc + "')."); 
    } 
  }
  
  public VerificationResult do_verify() {
    try {
      if (this.myOwner.doPass2().equals(VerificationResult.VR_OK)) {
        JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
        Method[] methods = jc.getMethods();
        if (this.methodNo >= methods.length)
          throw new InvalidMethodException("METHOD DOES NOT EXIST!"); 
        Method method = methods[this.methodNo];
        this.code = method.getCode();
        if (method.isAbstract() || method.isNative())
          return VerificationResult.VR_OK; 
        try {
          this.instructionList = new InstructionList(method.getCode().getCode());
        } catch (RuntimeException re) {
          return new VerificationResult(2, "Bad bytecode in the code array of the Code attribute of method '" + 
              tostring(method) + "'.");
        } 
        this.instructionList.setPositions(true);
        VerificationResult vr = VerificationResult.VR_OK;
        try {
          delayedPass2Checks();
        } catch (ClassConstraintException|org.apache.bcel.classfile.ClassFormatException cce) {
          vr = new VerificationResult(2, cce.getMessage());
          return vr;
        } 
        try {
          pass3StaticInstructionChecks();
          pass3StaticInstructionOperandsChecks();
        } catch (StaticCodeConstraintException|org.apache.bcel.classfile.ClassFormatException scce) {
          vr = new VerificationResult(2, scce.getMessage());
        } catch (ClassCastException cce) {
          vr = new VerificationResult(2, "Class Cast Exception: " + cce.getMessage());
        } 
        return vr;
      } 
      return VerificationResult.VR_NOTYET;
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public int getMethodNo() {
    return this.methodNo;
  }
  
  private void pass3StaticInstructionChecks() {
    if ((this.code.getCode()).length >= 65536)
      throw new StaticCodeInstructionConstraintException("Code array in code attribute '" + 
          tostring(this.code) + "' too big: must be smaller than " + 65536 + "65536 bytes."); 
    InstructionHandle ih = this.instructionList.getStart();
    while (ih != null) {
      Instruction i = ih.getInstruction();
      if (i instanceof org.apache.bcel.generic.IMPDEP1)
        throw new StaticCodeInstructionConstraintException("IMPDEP1 must not be in the code, it is an illegal instruction for _internal_ JVM use!"); 
      if (i instanceof org.apache.bcel.generic.IMPDEP2)
        throw new StaticCodeInstructionConstraintException("IMPDEP2 must not be in the code, it is an illegal instruction for _internal_ JVM use!"); 
      if (i instanceof org.apache.bcel.generic.BREAKPOINT)
        throw new StaticCodeInstructionConstraintException("BREAKPOINT must not be in the code, it is an illegal instruction for _internal_ JVM use!"); 
      ih = ih.getNext();
    } 
    Instruction last = this.instructionList.getEnd().getInstruction();
    if (!(last instanceof org.apache.bcel.generic.ReturnInstruction) && !(last instanceof org.apache.bcel.generic.RET) && !(last instanceof org.apache.bcel.generic.GotoInstruction) && !(last instanceof org.apache.bcel.generic.ATHROW))
      throw new StaticCodeInstructionConstraintException("Execution must not fall off the bottom of the code array. This constraint is enforced statically as some existing verifiers do - so it may be a false alarm if the last instruction is not reachable."); 
  }
  
  private void pass3StaticInstructionOperandsChecks() {
    try {
      ConstantPoolGen cpg = new ConstantPoolGen(Repository.lookupClass(this.myOwner.getClassName()).getConstantPool());
      InstOperandConstraintVisitor v = new InstOperandConstraintVisitor(this, cpg);
      InstructionHandle ih = this.instructionList.getStart();
      while (ih != null) {
        Instruction i = ih.getInstruction();
        if (i instanceof JsrInstruction) {
          InstructionHandle target = ((JsrInstruction)i).getTarget();
          if (target == this.instructionList.getStart())
            throw new StaticCodeInstructionOperandConstraintException("Due to JustIce's clear definition of subroutines, no JSR or JSR_W may have a top-level instruction (such as the very first instruction, which is targeted by instruction '" + 
                
                tostring(ih) + "' as its target."); 
          if (!(target.getInstruction() instanceof org.apache.bcel.generic.ASTORE))
            throw new StaticCodeInstructionOperandConstraintException("Due to JustIce's clear definition of subroutines, no JSR or JSR_W may target anything else than an ASTORE instruction. Instruction '" + 
                
                tostring(ih) + "' targets '" + tostring(target) + "'."); 
        } 
        ih.accept(v);
        ih = ih.getNext();
      } 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  protected String tostring(Object obj) {
    String ret;
    try {
      ret = obj.toString();
    } catch (RuntimeException e) {
      String s = obj.getClass().getName();
      s = s.substring(s.lastIndexOf(".") + 1);
      ret = "<<" + s + ">>";
    } 
    return ret;
  }
}
